import numpy as np

from keras.callbacks import Callback


class Pruning(Callback):
    """
    Set weights that fall between a certain bound to a certain value. Typically used to reduce model size by setting
    weights that are close to 0.0, to 0.0.

    Arguments:
        value: The value that pruned weights will be set to. (this should usually be 0.0)
        min_value: The lowest value a weight can be to be oeprated on
        max_value: The highest value a weight can be to be operated on
        cutoff: The number of epochs that Pruning will be performed
        biases: If true, Pruning will also be performed on biases
    """

    def __init__(self,
                 value: float = 0.0,
                 min_value: float = -.2,
                 max_value: float = .2,
                 cutoff: int = -1,
                 biases: bool = False):

        super(Pruning, self).__init__()

        self.value = value
        self.min_value = min_value
        self.max_value = max_value
        self.cutoff = cutoff
        self.biases = biases

        self.epochs_completed = 0

    def on_epoch_end(self, batch, logs=None):
        if self.cutoff > self.epochs_completed or self.cutoff == -1:
            weights = self.model.get_weights()[0].copy()
            biases = self.model.get_weights()[1].copy()

            for weight in np.nditer(weights, op_flags=['readwrite']):
                if self.min_value <= weight <= self.max_value:
                    weight = self.value

            if self.biases:
                for bias in np.nditer(biases, op_flags=['readwrite']):
                    if self.min_value <= bias <= self.max_value:
                        bias = self.value

            self.model.set_weights([weights, biases])



import numpy as np

import random

from keras.callbacks import Callback


class NeuroPlast(Callback):
    """
    Re-initialize weights that are 0 or close to 0, so that models can operate at 'full capacity'

    Arguments:
        min_value: The lowest value a weight/bias can be to be operated on
        max_value: The highest value a weight/bias can be to be operated on
        low_bound: When weights are re-initialized, this will be their lowest possible value
        high_bound: When weights are re-initialized, this will be their highest possible value
        biases: If true, NeuroPlast will also be performed on biases
    """

    def __init__(self,
                 min_value: float = -.02,
                 max_value: float = .02,
                 low_bound: float = -.05,
                 high_bound: float = .05,
                 biases: bool = False):

        super(NeuroPlast, self).__init__()

        self.min_value = min_value
        self.max_value = max_value
        self.low_bound = low_bound
        self.high_bound = high_bound
        self.biases = biases

    def on_epoch_end(self, batch, logs=None):
        weights = self.model.get_weights()[0].copy()
        biases = self.model.get_weights()[1].copy()

        for weight in np.nditer(weights, op_flags=['readwrite']):
            if self.min_value <= weight <= self.max_value:
                weight = random.uniform(self.low_bound, self.high_bound)

        if self.biases:
            for bias in np.nditer(biases, op_flags=['readwrite']):
                if self.min_value <= weight <= self.max_value:
                    bias = random.uniform(self.low_bound, self.high_bound)

        self.model.set_weights([weights, biases])



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='kerastroke',
    version='2.0.4',
    scripts=['./kerastroke/__init__.py'],
    author="Charles Averill",
    author_email="charlesaverill20@gmail.com",
    description="A suite of the generalization-improvement techniques Stroke, Pruning, and NeuroPlast",
    long_description=long_description,
    install_requires=['keras', 'numpy'],
    long_description_content_type="text/markdown",
    url="https://github.com/CharlesAverill/kerastroke/",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent"
    ]
)



import random
import numpy as np

from keras.callbacks import Callback


class Stroke(Callback):
    """
    Re-initialize a percentage of weights in a layer or model at the end of every epoch to improve generalization.
    
    Arguments:
        value: If set, re-initialized weights will be set to this value, rather than a random one
        low_bound: When weights are re-initialized, this will be their lowest possible value
        high_bound: When weights are re-initialized, this will be their highest possible value
        volatility_ratio: The percentage of weights that will be re-initialized after every epoch
        cutoff: The number of epochs that Stroke will be performed
        decay: Every epoch, v_ratio is multiplied by this number. decay can be greater than 1.0, but v_ratio will
               never exceed 1.0
        indeces: A list of integers specifying the indeces of layers that Stroke will be performed on, rather than
                 the model as a whole
        biases: If true, Stroke will also be performed on biases
    """

    def __init__(self,
                 value: float = None,
                 low_bound: float = -.05,
                 high_bound: float = .05,
                 volatility_ratio: float = .05,
                 cutoff: int = -1,
                 decay: float = None,
                 indeces=None,
                 biases: bool = False):

        super(Stroke, self).__init__()

        self.value = value
        self.low_bound = low_bound
        self.high_bound = high_bound
        assert volatility_ratio < 1.0, "volatility_ratio must be less than 1.0"
        self.v_ratio = volatility_ratio
        self.cutoff = cutoff
        self.decay = decay
        self.indeces = indeces
        self.biases = biases

        self.epochs_completed = 0

    def on_epoch_end(self, batch, logs=None):
        if self.cutoff > self.epochs_completed or self.cutoff == -1:
            if self.indeces is None:

                weights = self.model.get_weights()[0]
                biases = self.model.get_weights()[1]

                num_weights = len(np.nditer(weights))
                num_biases = len(np.nditer(biases))

                for stricken in range(int(num_weights * self.v_ratio)):
                    if self.value is None:
                        weights[tuple(map(lambda x: np.random.randint(0, x), weights.shape))] = random.uniform(
                                self.low_bound, self.high_bound)
                    else:
                        weights[tuple(map(lambda x: np.random.randint(0, x), weights.shape))] = self.value

                if self.biases:
                    for stricken in range(int(num_biases * self.v_ratio)):
                        if self.value is None:
                            biases[tuple(map(lambda x: np.random.randint(0, x), weights.shape))] = random.uniform(
                                    self.low_bound, self.high_bound)
                        else:
                            biases[tuple(map(lambda x: np.random.randint(0, x), weights.shape))] = self.value

                self.model.set_weights([weights, biases])

            else:
                for index in self.indeces:
                    weights = self.model.get_layer(index=index).get_weights()[0]
                    biases = self.model.get_layer(index=index).get_weights()[1]

                    num_weights = len(np.diter(weights))
                    num_biases = len(np.nditer(biases))

                    for stricken in range(int(num_weights * self.v_ratio)):
                        if self.value is None:
                            weights[tuple(map(lambda x: np.random.randint(0, x), weights.shape))] = random.uniform(
                                self.low_bound, self.high_bound)
                        else:
                            weights[tuple(map(lambda x: np.random.randint(0, x), weights.shape))] = self.value

                    if self.biases:
                        for stricken in range(int(num_biases * self.v_ratio)):
                            if self.value is None:
                                biases[tuple(map(lambda x: np.random.randint(0, x), weights.shape))] = random.uniform(
                                    self.low_bound, self.high_bound)
                            else:
                                biases[tuple(map(lambda x: np.random.randint(0, x), weights.shape))] = self.value

                    self.model.get_layer(index=index).set_weights([weights, biases])

        if self.decay is not None and .01 < self.v_ratio < 1.0:
            self.v_ratio = self.v_ratio * self.decay
